import sys
from getpass import getpass
import time
import re
from tapipy.tapis import Tapis
import socket
import os
import logging
from tapisObjectWrappers import Files, Apps, Pods, Systems, Neo4jCLI
import typing

try:
    from . import exceptions
    from . import socketOpts as SO
    from . import helpers
    from . import schemas
    from . import decorators
except:
    import exceptions
    import socketOpts as SO
    import helpers
    import schemas
    import decorators

class Server(SO.SocketOpts, helpers.OperationsHelper, decorators.DecoratorSetup, helpers.DynamicHelpUtility):
    def __init__(self, IP: str, PORT: int):
        # logger setup
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        stream_handler = logging.StreamHandler(stream=sys.stdout)

        file_handler = logging.FileHandler(
            r'logs.log', mode='w')
        stream_handler.setLevel(logging.INFO)
        file_handler.setLevel(logging.INFO)

        # set formats
        stream_format = logging.Formatter(
            '%(name)s - %(levelname)s - %(message)s')
        file_format = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        stream_handler.setFormatter(stream_format)
        file_handler.setFormatter(file_format)

        # add the handlers
        self.logger.addHandler(stream_handler)
        self.logger.addHandler(file_handler)

        self.logger.disabled = False

        # setting up socket server
        self.ip, self.port = IP, PORT
        self.sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.sock.bind((self.ip, self.port))
        self.sock.listen(1)
        self.connection = None  # initialize the connection variable
        self.end_time = time.time() + 300  # start the countdown on the timeout

        self.logger.info("Awaiting connection")

        self.pods = None
        self.systems = None
        self.files = None
        self.apps = None
        self.neo4j = None
        self.t = None
        self.url = None
        self.access_token = None
        self.username = None
        self.password = None

        self.accept(initial=True)  # connection returns the tapis object and user info

        # instantiate the subsystems
        self.logger.info('initialization complete')
        self.command_group_map = {
            'pods':self.pods,
            'systems':self.systems,
            'files':self.files,
            'apps':self.apps,
        }
        self.command_map = {
            'help':self.help,
            'whoami':self.pods.whoami,
            'exit':self.__exit,
            'shutdown':self.__shutdown,
            'neo4j':self.neo4j,
            'switch_service':self.tapis_init
        }
        help0, help1 = self.help_generation()
        self.help = dict(help0, **help1)

    @decorators.Auth
    def tapis_init(self, username: str, password: str, link: str) -> tuple[typing.Any, str, str] | None:  # link is the baseURL
        """
        @help: switch the connected tapis service
        """
        start = time.time()
        self.username = username
        self.password = password
        try:
            t = Tapis(base_url=link,
                    username=username,
                    password=password)
            t.get_tokens()
        except:
            raise exceptions.InvalidCredentialsReceived(function=self.tapis_init, cred_type="Tapis Auth")

        self.configure_decorators()
        # V3 Headers
        header_dat = {"X-Tapis-token": t.access_token.access_token,
                      "Content-Type": "application/json"}

        # Service URL
        url = f"{link}/v3"

        # create authenticator for tapis systems
        authenticator = t.access_token
        # extract the access token from the authenticator
        access_token = re.findall(
            r'(?<=access_token: )(.*)', str(authenticator))[0]
        
        if 'win' in sys.platform:
            os.system(f"set JWT={access_token}")
        else: # unix based
            os.system(f"export JWT={access_token}")

        self.pods = Pods(t, username, password, self.connection)
        self.systems = Systems(t, username, password, self.connection)
        self.files = Files(t, username, password, self.connection)
        self.apps = Apps(t, username, password, self.connection)
        self.neo4j = Neo4jCLI(t, username, password, self.connection)

        self.t = t
        self.url = url
        self.access_token = access_token

        self.logger.info(f"initiated in {time.time()-start}")

        return f"Successfully initialized tapis service on {self.url}"

    def accept(self, initial: bool=False):  # function to accept CLI connection to the server
        self.connection, ip_port = self.sock.accept()  # connection request is accepted
        self.logger.info("Received connection request")

        if initial:  # if this is the first time in the session that the cli is connecting
            startup_data = schemas.StartupData(initial = initial)
            self.json_send(startup_data.dict())
            self.logger.info("send the initial status update")

            # give the cli 3 attempts to provide authentication
            for attempt in range(1, 4):
                  # receive the username and password
                url: schemas.StartupData = self.schema_unpack().url
                try:
                # try intializing tapis with the supplied credentials
                    auth_request = schemas.AuthRequest()
                    self.json_send(auth_request.dict())
                    auth_data: schemas.AuthData = self.schema_unpack()
                    username, password = auth_data.username, auth_data.password

                    self.tapis_init(link=url, username=username, password=password)
                    # send to confirm to the CLI that authentication succeeded
                    self.logger.info("Verification success")
                    break
                except Exception as e:
                    # send failure message to CLI
                    login_failure_data = schemas.ResponseData(response_message = (str(e), attempt))
                    self.json_send(login_failure_data.dict())
                    self.logger.warning(f"Verification failure, {e}")
                    if attempt == 3:  # If there have been 3 login attempts
                        self.logger.error(
                            "Attempted verification too many times. Exiting")
                        os._exit(0)  # shutdown the server
                    continue
        else:
            self.configure_decorators()
        startup_result = schemas.StartupData(initial = initial, username = self.username, url = self.url)
        self.logger.info("Connection success")
        self.json_send(startup_result.dict())
        self.logger.info("Final connection data sent")

    def __exit(self):
        """
        @help: exit the CLI without shutting down the service
        """
        raise exceptions.Exit
    
    def __shutdown(self):
        """
        @help: exit the CLI and shutdown the service
        """
        self.logger.info("Shutdown initiated")
        raise exceptions.Shutdown

    def timeout_handler(self):  # handle timeouts
        if time.time() > self.end_time:  # if the time exceeds the timeout time
            raise exceptions.TimeoutError
    
    def format_help(self, command: dict):
        return f"Command: {command['command_name']}\nDescription:{command['description']}\n{command['syntax']}\n"


    def help(self, command: str):
        """
        @help: returns help information. To get specific help information for tapis services, you can run <service> -c help
        """
        if command in self.help:
            return self.help[command]
        return self.help

    def run_command(self, command_data: dict):  # process and run commands
        command_group = command_data['command_group']
        if command_group in self.command_group_map:
            command_group = self.command_group_map[command_group]
            return command_group(**command_data)
        elif command_group in self.command_map:
            command = self.command_map[command_group]
            command_data = self.filter_kwargs(command, command_data)
            if command_data:
                return command(**command_data)
            return command()
        else:
            raise exceptions.CommandNotFoundError(command_group)

    def main(self):
        while True: 
            try:
                message = self.schema_unpack()  
                self.timeout_handler()  
                kwargs, exit_status = message.kwargs, message.exit_status
                result = self.run_command(kwargs)
                response = schemas.ResponseData(response_message = result)
                self.end_time = time.time() + 300 
                self.json_send(response.dict()) 
                self.logger.info(message)
                if exit_status == 1:
                    self.__exit()
            except (exceptions.CommandNotFoundError, exceptions.NoConfirmationError, exceptions.InvalidCredentialsReceived) as e:
                error_response = schemas.ResponseData(response_message = str(e))
                self.json_send(error_response.dict())
            except (exceptions.TimeoutError, exceptions.Shutdown) as e:
                error_response = schemas.ResponseData(response_message = str(e), exit_status=1)
                self.json_send(error_response.dict())
                sys.exit(0)
            except exceptions.Exit as e:
                self.logger.info("user exit initiated")
                error_response = schemas.ResponseData(response_message = str(e), exit_status=1)
                self.json_send(error_response.dict())
                self.connection.close()  # close the connection
                self.accept()  # wait for CLI to reconnect



if __name__ == '__main__':
    server = Server(socket.gethostbyname(socket.gethostname()), 30000)
    server.main()



import typing
import socket
import sys
import time
from functools import update_wrapper, partial
try:
    from . import helpers
    from . import schemas
    from . import socketOpts
    from . import exceptions
except:
    import helpers
    import schemas
    import socketOpts
    import exceptions


class BaseRequirementDecorator(socketOpts.SocketOpts, helpers.OperationsHelper):
    connection: typing.Optional[socket.socket] = None
    username: typing.Optional[str] = None
    password: typing.Optional[str] = None
    def __init__(self, func: typing.Callable):
        update_wrapper(self, func)
        self.function = func
        self.__code__ = func.__code__
        self.__doc__ = func.__doc__
        self.__name__ = func.__name__
    
    def __get__(self, obj, objtype): 
        """Support instance methods."""
        part = partial(self.__call__, obj)
        part.__code__ = self.__code__
        part.__doc__ = self.__doc__
        part.__name__ = self.__name__
        return part
    
    def __repr__(self):
        return str(self.function)
    
    def __str__(self):
        return str(self.function)


class RequiresForm(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        fields = list(helpers.get_parameters(self.function))
        for key, value in kwargs.items():
            if value or value == False:
                fields.remove(key)
        if not fields:
            raise AttributeError(f"The decorated function {self.function} has no parameters.")
        form_request = schemas.FormRequest(arguments_list=fields)
        self.json_send_explicit(BaseRequirementDecorator.connection, form_request.dict())
        filled_form: schemas.FormResponse = self.schema_unpack_explicit(self.connection).arguments_list
        for key, value in filled_form.items():
            kwargs[key] = value

        return self.function(obj, **kwargs)


class RequiresExpression(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        fields = list(helpers.get_parameters(self.function))
        if 'expression' not in fields:
            raise AttributeError(f"The function {self.function} does not contain an 'expression' parameter")
        form_request = schemas.FormRequest(arguments_list=[])
        self.json_send_explicit(BaseRequirementDecorator.connection, form_request.dict())
        filled_form: schemas.FormResponse = self.schema_unpack()
        kwargs['expression'] = filled_form.arguments_list

        return self.function(obj, **kwargs)
    

class SecureInput(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        fields = list(helpers.get_parameters(self.function))
        if 'password' in fields:
            secure_input_request = schemas.AuthRequest(secure_input=True)
            self.json_send_explicit(BaseRequirementDecorator.connection, secure_input_request.dict())
            secure_input_data: schemas.AuthData = self.schema_unpack_explicit(self.connection)
            kwargs['password'] = secure_input_data.password
            return self.function(obj, **kwargs)
        raise AttributeError(f"The function {self.function} does not contain a 'password' parameter")


class Auth(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        if self.function.__name__ == 'tapis_init' and kwargs['username'] and kwargs['password']:
            return self.function(obj, **kwargs)
        fields = list(helpers.get_parameters(self.function))
        auth_request = schemas.AuthRequest()
        self.json_send_explicit(BaseRequirementDecorator.connection, auth_request.dict())
        auth_data: schemas.AuthData = self.schema_unpack_explicit(self.connection)
        if 'username' in fields and 'password' in fields:
            kwargs['username'], kwargs['password'] = auth_data.username, auth_data.password
            return self.function(obj, **kwargs)
        username, password = auth_data.username, auth_data.password
        if username != BaseRequirementDecorator.username:
            raise exceptions.InvalidCredentialsReceived(self.function, 'username')
        elif password != BaseRequirementDecorator.password:    
            raise exceptions.InvalidCredentialsReceived(self.function, 'password')

        return self.function(obj, **kwargs)


class NeedsConfirmation(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        confirmation_request = schemas.ConfirmationRequest(message=f"You requested to {self.function.__name__}. Please confirm (y/n)")
        self.json_send_explicit(BaseRequirementDecorator.connection, confirmation_request.dict())
        confirmation_reply: schemas.ResponseData = self.schema_unpack_explicit(self.connection)
        confirmed = confirmation_reply.response_message
        if not confirmed:
            raise exceptions.NoConfirmationError(self.function)
        return self.function(obj, **kwargs)
    
class TestDecorator(BaseRequirementDecorator):
    def __call__(self, obj, *args, **kwargs):
        print(BaseRequirementDecorator.connection)
        return self.function(obj, **kwargs)
    
class DecoratorSetup:
    def configure_decorators(self):
        BaseRequirementDecorator.connection = self.connection
        BaseRequirementDecorator.username = self.username
        BaseRequirementDecorator.password = self.password
    

class AnimatedLoading:
    def __init__(self, func: typing.Callable):
        update_wrapper(self, func)
        self.function = func
        self.__code__ = func.__code__
        self.animation_frames = ['|','/','-','\\']

    def __get__(self, obj, objtype):
        """Support instance methods."""
        return partial(self.__call__, obj)

    def __repr__(self):
        return self.function
    
    def __str__(self):
        return str(self.function)
    
    def animation(self):
        while True:
            for frame in self.animation_frames:
                sys.stdout.write(f'\rloading ' + frame)
                sys.stdout.flush()
                time.sleep(0.5)
    
    def __call__(self, obj, *args, **kwargs):
        animation_thread = helpers.KillableThread(target=self.animation)
        animation_thread.start()
        result = self.function(obj, *args, **kwargs)
        animation_thread.kill()
        return result


if __name__ == "__main__":
    class Silly(DecoratorSetup):
        def __init__(self):
            self.connection = "x"
            self.username = "y"
            self.password = "v"
            self.configure_decorators()

        @TestDecorator
        def sillify(self, x):
            return "THIS WORKED"
        
    zeugmizer = Silly()
    print(zeugmizer.sillify(x="zed"))


from typing import Any, Optional
from pydantic import BaseModel


class CommandData(BaseModel):
    schema_type: str = 'CommandData'
    kwargs: Optional[dict]
    expression: Optional[str] 
    exit_status: int = 0


class AuthData(BaseModel):
    schema_type: str = 'AuthData'
    username: Optional[str]
    password: Optional[str]


class StartupData(BaseModel):
    schema_type: str = 'StartupData'
    initial: bool = False
    username: Optional[str]
    url: Optional[str]


class ResponseData(BaseModel):
    schema_type: str = 'ResponseData'
    response_message: Any
    exit_status: int = 0


class FormRequest(BaseModel):
    schema_type: str = 'FormRequest'
    arguments_list: list


class FormResponse(BaseModel):
    schema_type: str = 'FormResponse'
    arguments_list: dict | str


class AuthRequest(BaseModel):
    schema_type: str = 'AuthRequest'
    secure_input: bool = False


class ConfirmationRequest(BaseModel):
    schema_type: str = 'ConfirmationRequest'
    message: str



import json
import pyperclip
import tapipy
from tapipy.tapis import TapisResult
from py2neo import Graph
import typing
import os
try:
    from . import helpers
    from . import decorators
except:
    import helpers 
    import decorators


__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))
server_path = os.path.join(__location__, 'server.py')


class tapisObject(helpers.OperationsHelper, decorators.DecoratorSetup, helpers.DynamicHelpUtility):
    def __init__(self, tapis_instance, username, password, connection, command_map=None):
        self.t = tapis_instance
        self.username = username
        self.password = password
        self.connection = connection

        self.command_map = command_map
        
        if self.command_map:
            self.help = self.help_generation()

    def __call__(self, **kwargs):
        try:
            command = self.command_map[kwargs['command']]
            kwargs = self.filter_kwargs(command, kwargs)
            result = command(**kwargs)
            if type(result) == TapisResult:
                return str(result)
            return result
        except (tapipy.errors.NotFoundError, tapipy.errors.BadRequestError, tapipy.errors.BaseTapyException) as e:
            return str(e)
        except KeyError:
            return self.help
    
    def help(self, name: typing.Optional[str]):
        """
        @help: get help information for the command group
        """
        if name:
            return self.help[name]
        return self.help
    

class TapisQuery(tapisObject):
    def __init__(self, tapis_object, uname, pword, connection):
        super().__init__(tapis_object, uname, pword, connection)
        self.t = tapis_object
        self.__code__ = self.query.__code__

    def __call__(self, **kwargs):
        try:
            kwargs = self.filter_kwargs(self.query, kwargs)
            result = self.query(**kwargs)
            return result
        except (tapipy.errors.NotFoundError, tapipy.errors.BadRequestError, tapipy.errors.BaseTapyException) as e:
            return str(e)


class Systems(tapisObject):
    """
    @help: Access Tapis systems through the connected service
    """
    def __init__(self, tapis_instance, username, password, connection):
        command_map = {
            'get_systems':self.get_systems,
            'get_system_info':self.get_system_info,
            'create_system':self.create_system,
            'set_credentials':self.system_credential_upload,
            'set_password':self.system_password_set,
            'delete_system':self.delete_system,
            'help':self.help
        }
        super().__init__(tapis_instance, username, password, connection, command_map=command_map)

    def return_formatter(self, info):
        return f"id: {info.id}\nhost: {info.host}\n\n"

    def __keygen(self):
        local_files = os.listdir(__location__)
        if "id_rsa" not in local_files or "id_rsa.pub" not in local_files:
            os.system(f'ssh-keygen -q -m PEM -f {__location__}\\id_rsa -N ""')
            with open(f"{__location__}\\id_rsa", 'r') as f:
                formatted_key = ""
                for line in f.readlines()[1:-1]:
                    formatted_key += line.strip()

            with open(f"{__location__}\\id_rsa", 'w') as f:
                f.write(formatted_key)

    def get_systems(self, verbose: bool):
        """
        @help: Gets and returns the list of systems the current Tapis service and account have access to
        @doc: this is an example of the doc segment of the docstring. not included in help message
        """
        systems = self.t.systems.getSystems()
        if systems and verbose:
            return str(systems)
        systems = [self.return_formatter(system) for system in systems]
        systems_string = ''
        for system in systems:
            systems_string += system
        return systems_string

    def get_system_info(self, verbose: bool, id:str): # get information about a system given its ID
        """
        @help: get information on a selected system
        """
        system_info = self.t.systems.getSystem(systemId=id)
        if verbose:
            return str(system_info)
        return self.return_formatter(system_info)
    
    def create_system(self, file: str) -> str: # create a tapius system. Takes a path to a json file with all system information, as well as an ID
        """
        @help: create a system. Must have a properly configured system file.
        see the template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/system-config.json
        this command will automatically create and upload the ssh keys
        """
        self.__keygen()
        with open(file, 'r') as f:
            system = json.loads(f.read())
        return_value = self.t.systems.createSystem(**system)
        self.system_credential_upload(id=system['id'], file=f"{__location__}\\id_rsa,{__location__}\\id_rsa.pub")
        return str(return_value)
    
    def system_credential_upload(self, id: str, file: str) -> str: # upload key credentials for the system
        """
        @help: upload system credentials to a system. Must generate keys first using 'ssh-keygen -m PEM -f id_rsa', and format with, 'awk -v ORS='\\n' '1' <private_key_name>
        file argument must contain the path to the private and public keys respectively, separated by a ','
        """
        with open(file.split(",")[0], 'r') as f:
            private_key = f.read()

        with open(file.split(",")[1], 'r') as f:
            public_key = f.read()

        cred_return_value = self.t.systems.createUserCredential(systemId=id,
                            userName=self.username,
                            privateKey=private_key,
                            publicKey=public_key)

        return str(cred_return_value)

    @decorators.SecureInput
    def system_password_set(self, id: str, password: str) -> str: # set the password for a system
        """
        @help: set a system password
        """
        try:
            password_return_value = self.t.systems.createUserCredential(systemId=id, # will put this in a getpass later
                                userName=self.username,
                                password=password)
            return str(password_return_value)
        except Exception as e:
            raise Exception(f"{e}\nTry running set_credentials if the problem persists")

    @decorators.NeedsConfirmation
    def delete_system(self, id: str) -> str:
        """
        @help: delete the selected system
        """
        return_value = self.t.systems.deleteSystem(systemId=id)
        return return_value


class Neo4jCLI(TapisQuery):
    """
    @help: integrated CLI to interface with Neo4j pods
    """
    @decorators.RequiresExpression
    def query(self, id: str, expression: str) -> str: # function to submit queries to a Neo4j knowledge graph
        uname, pword = self.t.pods.get_pod_credentials(pod_id=id).user_username, self.t.pods.get_pod_credentials(pod_id=id).user_password
        graph = Graph(f"bolt+ssc://{id}.pods.icicle.tapis.io:443", auth=(uname, pword), secure=True, verify=True)

        try:
            return_value = graph.run(expression)
            print(type(return_value))
            if str(return_value) == '(No data)' and 'create' in expression.lower(): # if no data is returned (mostly if something is created) then just say 'success'
                return f'[+][{id}@pods.icicle.tapis.io:443] Success'
            elif str(return_value) == '(No data)':
                return f'[-][{id}@pods.icicle.tapis.io:443] KG is empty'

            print(return_value)
            print(type(return_value))
            return str(f'[+][{id}] {return_value}')
        except Exception as e:
            return str(e)


class Pods(tapisObject):
    """
    @help: Access Tapis pods through the connected service
    """
    def __init__(self, tapis_instance, username, password, connection):
        command_map = {
                'get_pods':self.get_pods,
                'create_pod':self.create_pod,
                'restart_pod':self.restart_pod,
                'delete_pod':self.delete_pod,
                'set_pod_perms':self.set_pod_perms,
                'stop_pod':self.stop_pod,
                'delete_pod_perms':self.delete_pod_perms,
                'get_perms':self.get_perms,
                'copy_pod_password':self.copy_pod_password,
                'get_logs':self.get_pod_logs,
                'help':self.help
            }
        super().__init__(tapis_instance, username, password, connection, command_map=command_map)

    def return_formatter(self, info):
        return f"Pod ID: {info.pod_id}\nPod Template: {info.pod_template}\nStatus: {info.status_requested}\n\n"

    def get_pods(self, verbose: bool) -> str: 
        """
        @help: return a list of pods the current tapis instance has access to
        @doc: notes
        """
        pods_list = self.t.pods.get_pods()
        if verbose:
            return str(pods_list)
        pods_list = [self.return_formatter(pod) for pod in pods_list]
        pods_string = ""
        for pod in pods_list:
            pods_string += str(pod)
        return pods_string
    
    def whoami(self, verbose: bool) -> str:
        """
        @help: returns the username of the current user
        """
        user_info = self.t.authenticator.get_userinfo()
        if verbose:
            return str(user_info)
        return user_info.username

    @decorators.RequiresForm
    def create_pod(self, description: str, id: str, template: str, verbose: bool) -> str:
        """
        @help: create a new pod on the selected Tapis service
        """
        pod_information = self.t.pods.create_pod(pod_id=id, pod_template=template, description=description)
        if verbose:
            return str(pod_information)
        return pod_information

    @decorators.NeedsConfirmation
    def restart_pod(self, id: str, verbose: bool) -> str:
        """
        @help: initiate a pod restart
        """
        return_information = self.t.pods.restart_pod(pod_id=id)
        if verbose:
            return str(return_information)
        return return_information
    
    @decorators.NeedsConfirmation
    def stop_pod(self, id: str):
        """
        @help: stop a pod's operations
        """
        return_information = self.t.pods.stop_pod(pod_id=id)
        return return_information
        
    @decorators.NeedsConfirmation
    def delete_pod(self, id: str, verbose: bool) -> str: 
        """
        @help: delete select pod
        """
        return_information = self.t.pods.delete_pod(pod_id=id)
        if verbose:
            return str(return_information)
        return return_information

    def set_pod_perms(self, id: str, username: str, level: str) -> str: # set pod permissions, given a pod id, user, and permission level
        """
        @help: set the permissions for the pod selected
        """
        return_information = self.t.pods.set_pod_permission(pod_id=id, user=username, level=level)
        return str(return_information)
    
    @decorators.NeedsConfirmation
    def delete_pod_perms(self, id: str, username: str) -> str: # take away someones perms if they are being malicious, or something
        """
        @help: delete the selected pod from the pods service you are connected to
        """
        return_information = self.t.pods.delete_pod_perms(pod_id=id, user=username)
        return str(return_information)

    def get_perms(self, id: str) -> str: # return a list of permissions on a given pod
        """
        @help: get the permissions list for the selected pod
        """
        return_information = self.t.pods.get_pod_permissions(pod_id=id)
        return str(return_information)

    @decorators.Auth
    def copy_pod_password(self, id: str) -> str: # copies the pod password to clipboard so that the user can access the pod via the neo4j desktop app. Maybe a security risk? not as bad as printing passwords out!
        """
        @help: copy the pod password to the clipboard
        """
        password = self.t.pods.get_pod_credentials(pod_id=id).user_password
        pyperclip.copy(password)
        password = None
        return 'copied to clipboard'
    
    def get_pod_logs(self, id: str, file=None):
        """
        @help: retrieve the logs of an active pod and either print them to the console, or write them to the specified file
        """
        logs = self.t.pods.get_pod_logs(pod_id=id)
        if file:
            with open(file, 'w') as f:
                f.write(logs)
            return f"Log saved at {file}"
        return logs


class Files(tapisObject):
    """
    @help: Access Tapis files through the connected service
    """
    def __init__(self, tapis_instance, username, password, connection):
        command_map = {
            'list_files':self.list_files,
            'upload':self.upload,
            'download':self.download,
            'help':self.help
        }
        super().__init__(tapis_instance, username, password, connection, command_map=command_map)

    def return_formatter(self, info):
        return f"name: {info.name}\ngroup: {info.group}\npath: {info.path}\n"

    def list_files(self, verbose: bool, id: str, file: str) -> str: # lists files available on a tapis account
        """
        @help: list the files on a system 
        """
        file_list = self.t.files.listFiles(systemId=id, path=file)
        if verbose:
            return str(file_list)
        file_list = [self.return_formatter(f) for f in file_list]
        return str(file_list)

    def upload(self, file: str, id: str) -> str: # upload a file from local to remote using tapis. Takes source and destination paths
        """
        @help: upload a file to the system 
        the source and destination files must both be in the file argument, respectively, separated by a comma
        """
        source = file.split(",")[0]
        destination = file.split(",")[1]
        self.t.upload(system_id=id,
                source_file_path=source,
                dest_file_path=destination)
        return f'successfully uploaded {source} to {destination}'
            
    def download(self, file: str, id: str) -> str: # download a remote file using tapis, operates basically the same as upload
        """
        @help: download a file from the system
        the source and destination files must both be in the file argument, respectively, separated by a comma
        """
        source = file.split(",")[0]
        destination = file.split(",")[1]
        file_info = self.t.files.getContents(systemId=id,
                            path=source)

        file_info = file_info.decode('utf-8')
        with open(destination, 'w') as f:
            f.write(file_info)
        return f'successfully downloaded {source} to {destination}'


class Apps(tapisObject):
    """
    @help: Access Tapis systems through the connected service
    """
    def __init__(self, tapis_instance, username, password, connection):
        command_map = {
            'create_app':self.create_app,
            'get_apps':self.get_apps,
            'delete_app':self.delete_app,
            'get_app_info':self.get_app,
            'run_app':self.run_job,
            'get_app_status':self.get_job_status,
            'download_app_results':self.download_job_output,
            'help':self.help,
        }
        super().__init__(tapis_instance, username, password, connection, command_map=command_map)

    def create_app(self, file: str) -> str: # create a tapis app taking a json descriptor file path
        """
        @help: create an app. You must have a properly configured app config file. 
        See a template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/app-config.json
        """
        with open(file, 'r') as f:
            app_def = json.loads(f.read())
        url = self.t.apps.createAppVersion(**app_def)
        return f"App created successfully\nID: {app_def['id']}\nVersion: {app_def['version']}\nURL: {url}\n"

    def get_apps(self) -> str:
        """
        @help: get all the apps on a system
        """
        apps = self.t.apps.getApps()
        return str(apps)

    @decorators.NeedsConfirmation
    def delete_app(self, id: str, version: str) -> str:
        """
        @help: delete the selected app
        """
        return_value = self.t.apps.deleteApp(appId=id, appVersion=version)
        return str(return_value)

    def get_app(self, verbose: bool, id: str, version: str)-> None | str: # returns app information with an id and version as arguments
        """
        @help: return selected app information
        """
        app = self.t.apps.getApp(appId=id, appVersion=version)
        if verbose:
            return str(app)
        return None

    def run_job(self, file: str)->str: # run a job using an app. Takes a job descriptor json file path
        """
        @help: run a job from an app on a system. You must have a properly configured job config file. 
        See a template at https://github.com/sdsc-hpc-training-org/hello_icicle_auth_clients/blob/main/icicle_rel_04_2023/CLI/TapisCL-ICICLE/tapis-config-files/job-config.json
        """
        with open(file, 'r') as f:
            app = json.loads(f.read())
        job = self.t.jobs.submitJob(**app)
        return str(job.uuid)

    def get_job_status(self, uuid: str)->str: # return a job status with its Uuid
        """
        @help: get the status of a job
        """
        job_status = self.t.jobs.getJobStatus(jobUuid=uuid)
        return str(job_status)

    def download_job_output(self, uuid: str, file: str)->str: # download the output of a job with its Uuid
        """
        @help: download a job output from the system 
        """
        jobs_output = self.t.jobs.getJobOutputDownload(jobUuid=uuid, outputPath='tapisjob.out')
        with open(file, 'w') as f:
            f.write(jobs_output)
        return f"Successfully downloaded job output to {file}"
    